/*
 * pts.h:	P(resentation) T(ime) S(tamps) handling
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * Copyright (C) 2003 Werner Fink, <werner@suse.de>
 *
 * Changes and Additional Code:
 *
 * Copyright (C) 2003 Sven Goethel, <sven@jausoft.com>
 */

#ifndef _PTS_H
#define _PTS_H

#include <unistd.h>
#include <stdint.h>
#include <sys/time.h>
#include "types.h"
#include "any_decoder.h"

# define debug_pts(args...)
// # define debug_pts(args...)	fprintf(stderr, args)

enum ePTSstate { eOK = 0, eMISSED = -1, eDOSKIP = 1 };

class cPTS {
private:
    // to hold the current state
    uint_32 pts, stc;
    uint_32 l_pts, l_stc;
    uint_32 start_pts, start_stc;
    struct timeval ts;

public:
    static inline uint_32 PTSinMS(const uchar *pts)
    {
	return (uint_32)
	((   ((((uint64_t)pts[0]) & 0x0E) << 29)
	   | (( (uint64_t)pts[1])         << 22)
	   | ((((uint64_t)pts[2]) & 0xFE) << 14)
	   | (( (uint64_t)pts[3])         <<  7)
	   | ((((uint64_t)pts[4]) & 0xFE) >>  1))/90);
    }

/*

    sender (VDR)                    <->          receiver (spdif)
        |                                            |
     forward credits (ms)            ->           use credits if not expired
   + current time stamp for expire                if no data for credits repeat last frame

                                    <-            if expired put backward credits
                                                  of repeated frames

     if expired get backward credits
     for skipping                    ->           do skipping 
 */

    cPTS(void) 
    : pts(0), stc(0), 
      l_pts(0), l_stc(0), 
      start_pts(0), start_stc(0)
    { };

    inline void mark(const uchar *ptsdata)
    {
	pts = PTSinMS(ptsdata);
    };
    inline void clear()
    {
	pts = 0;
	stc = 0;
	l_pts = 0;
	l_stc = 0;
	start_pts=0;
	start_stc=0;
    }
    inline uint_32 getTimeDiff()
    {
    	uint_32 td;
	struct timeval cur, diff;
	gettimeofday(&cur, 0);
	timersub(&cur, &ts, &diff);
	td = (diff.tv_sec*1000+diff.tv_usec/1000);
	gettimeofday(&ts, 0);
	return td;
    };
    inline uint_32 delay()
    {
	return stc - pts;
    };
    inline uint_32 getSTC() { return stc; }

    inline uint_32 getPTS() { return pts; }

    
    inline int process( const uchar *ptsdata, uint_32 stc, int sleep_threshold,
    		         AnyDecoder * AnyDecOptional=NULL )
    {
	if ( ptsdata==NULL) 
		return 0;

	if ( stc == 0 ) 
		return 0;

	mark(ptsdata); //  pts setup

	return process ( stc, sleep_threshold, AnyDecOptional);

    }

    inline int process( uint_32 stc, int sleep_threshold , AnyDecoder * AnyDecOptional=NULL )
    {
	int     delay_pts;
	// int     delay_pts_d;
	uint_32 u_sleep;

	this->stc = stc;

	if ( AnyDecOptional != NULL )
		AnyDecOptional->setDelay(0);

	// if cleared -> setup
	if ( start_pts==0 || start_stc==0 ) {
		start_stc=stc;
		start_pts=pts;
		l_stc=stc;
		l_pts=pts;
		debug_pts("PTS mark reset start stc, pts: %4lums, %4lums\n", stc, pts);
	}

/**
 	if((stc-start_stc)>(pts-start_pts)) {
		delay_pts_d = (int)((stc-start_stc)-(pts-start_pts)) ;	 // late
	} else {
		delay_pts_d = (int)((pts-start_pts)-(stc-start_stc)) * -1; // early
	}
 */
 	if(stc>pts) {
		delay_pts = (int)(stc-pts) ;	 // late
	} else {
		delay_pts = (int)(pts-stc) * -1; // early
	}

/**
	debug_pts("PTSDIFF stc-pts: %9lums (d%4lums) - %9lums (d%4lums) = %4dms; %6lums - %6lums = %4dms; ", 
		stc, stc-l_stc, pts, pts-l_pts, delay_pts,
		stc-start_stc, pts-start_pts, delay_pts_d
	);
 */
	debug_pts("PTSDIFF stc-pts: %9lums (d%4lums) - %9lums (d%4lums) = %4dms; ", 
		stc, stc-l_stc, pts, pts-l_pts, delay_pts
	);

	if(abs(delay_pts)>5000) {
		delay_pts=0;
	}

	if ( delay_pts > 5 ) {
		if ( AnyDecOptional != NULL ) {
			AnyDecOptional->setDelay((uint_32)delay_pts);
			debug_pts(" mp2dec delay  %4dms", delay_pts);
		}
	}  
	else if ( sleep_threshold>0 && delay_pts < sleep_threshold*-1 ) {
		u_sleep=(delay_pts+sleep_threshold)*-1000;

    		(void) getTimeDiff();

		usleep ( u_sleep );

		debug_pts(" sleep %4lums (%4lums)", 
			u_sleep/1000, getTimeDiff());
	}
	debug_pts("\n");
	l_stc=stc;
	l_pts=pts;

	return delay_pts;
    }
};

#endif

